/************************************************************************************************
 * Copyright  (c) 2011, 2012, 2013 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

/****************************************************
Test Bench

Author: Udit Dhawan
	IC Group, University of Pennsylvania

*****************************************************/


package Test;

import Wrapper::*;
import Vector::*;

`include "dmhc.cfg"

module mkTest(Empty);

	WrapperIfc dmhc_wrapper <- mkWrapper;

	Reg#(Bit#(64)) cycle <- mkReg(0);
	Reg#(Bit#(64)) read_cycle <- mkReg(0);
	Reg#(Bool) start <- mkReg(False);
	Reg#(Bool) success <- mkReg(False);
	Reg#(Bit#(8)) count <- mkReg(0);
	Reg#(Bool) miss <- mkReg(False);

	Vector#(5, Bit#(64)) keys;
	keys[0]=100; keys[1]=200; keys[2]=300; keys[3]=400; keys[4]=500;
	Vector#(5, Bit#(64)) values;
	values[0]=101; values[1]=201; values[2]=301; values[3]=401; values[4]=501;

	//initialized a 1024-entry dMHC 
	//-- it should take 2048 cycles to initialize the memories

	rule inc_cycle;
		if(cycle == 2049) begin
			start <= True;
		end
		cycle <= cycle + 1;
	endrule

	rule this_is_end (count == 10);
		$finish();
	endrule

	rule give_key(start && (cycle%1000 == 0));
		$display("[%d]: looking up %d", $time, keys[count%5]);
		let tmp <- dmhc_wrapper.lookup_key(keys[count%5]);

		if(!tmp) begin
			$display("[%d]: key lookup failed", $time);
		end
		else begin
			success <= True;
		`ifdef FAST_VALUE
			read_cycle <= cycle+1;
		`elsif TWO_LEVEL
			read_cycle <= cycle+2;
		`endif
		end
	endrule

`ifdef FAST_VALUE
	rule get_value(success && (cycle == read_cycle));
		let value = dmhc_wrapper.get_value;
		$display("[%d]: value: %d", $time, value); 
	endrule
	rule get_hit(success && (cycle == read_cycle+1));
		let hit = dmhc_wrapper.is_hit;
		if(hit) begin
			$display("[%d]: hit", $time);
			count <= count + 1;
		end
		else begin
			$display("[%d]: miss", $time);
			miss <= True;
		end
		success <= False;
	endrule
`endif

`ifdef TWO_LEVEL
	rule get_result(success && (cycle == read_cycle));
		let hit = dmhc_wrapper.is_hit;
		if(hit) begin
			$display("[%d]: hit", $time);
			let value = dmhc_wrapper.get_value;
			$display("[%d]: value: %d", $time, value); 
			count <= count + 1;
		end
		else begin
			$display("[%d]: miss", $time);
			miss <= True;
		end
		success <= False;
	endrule
`endif

	rule insert_keyvalue(miss);
		$display("[%d]: inserting (%d,%d)", $time, keys[count%5], values[count%5]);
		dmhc_wrapper.new_key_value(keys[count%5], values[count%5]);
		miss <= False;
		count <= count + 1;
	endrule

endmodule: mkTest

endpackage
