/************************************************************************************************
 * Copyright  (c) 2011, 2012, 2013 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

/****************************************************
Top level wrapper for the dMHC module

Author: Udit Dhawan
	IC Group, University of Pennsylvania

*****************************************************/


package Wrapper;

import Vector::*;
import DMHC::*;

interface WrapperIfc;
	//method Action flush();
	method ActionValue#(Bool) lookup_key(Key k);
        method Value get_value();
	method Bool is_hit();
	method Action new_key_value(Key k, Value v);
endinterface: WrapperIfc


`include "dmhc.cfg"

typedef Bit#(`Wk) Key;
typedef Bit#(`Wv) Value;

(* synthesize *)
module mkWrapper(WrapperIfc);

	// dMHC instantiation
	DMHCIfc#(`N, `K, `C, `Wk, `Wv) dmhc <- mkDMHC(`N, `K, `C, `Wk, `Wv);

	//method Action flush();
	//	dmhc.flush;
	//endmethod

	method ActionValue#(Bool) lookup_key(Key k);
		let tmp <- dmhc.lookup_key(unpack(pack(k)));
		return tmp;
	endmethod

	method Value get_value();
		return unpack(pack(dmhc.get_value));
	endmethod

	method Bool is_hit();
		return dmhc.is_hit;
	endmethod

	method Action new_key_value(Key k, Value v);
		dmhc.new_key_value(unpack(pack(k)), unpack(pack(v)));
	endmethod

endmodule: mkWrapper

endpackage
